"""This is the entry-point to the kernreg package.

Include only imports which should be available using

.. code-block::

    import kernreg as kr

    kr.<func>
"""
from kernreg.bandwidth_selection import get_bandwidth_rsc  # noqa: F401
from kernreg.smooth import locpoly  # noqa: F401
from kernreg.utils import get_example_data, sort_by_x  # noqa: F401
from kernreg.visualize import plot  # noqa: F401


# Automatic package version update
try:
    from importlib.metadata import version, PackageNotFoundError  # type: ignore
except ImportError:  # pragma: no cover
    from importlib_metadata import version, PackageNotFoundError  # type: ignore


try:
    __version__ = version(__name__)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"



"""Visualize results."""
from typing import Optional, Union

import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from kernreg.smooth import Result


def plot(
    x_raw: Union[np.ndarray, pd.Series],
    y_raw: Union[np.ndarray, pd.Series],
    rslt: Result,
    save_as: Optional[str] = None,
) -> None:
    """Plot fitted curve against raw data.

    Arguments:
        x_raw: Raw predictor data.
        y_raw: Raw response data.
        rslt: Result dictionary from :func:`~kernreg.smooth.locpoly`
        save_as: File path and type (e.g. '.png', '.jpg') to save the image.
            If ``None``, the image is not saved.

    """
    ax = plt.figure(figsize=(17.5, 10)).add_subplot(111)

    if isinstance(x_raw, pd.Series) and isinstance(y_raw, pd.Series):
        xlabel, ylabel = x_raw.name, y_raw.name
    else:
        xlabel, ylabel = "Predictor", "Response"

    ax.set_xlabel(xlabel, fontsize=20)
    ax.set_ylabel(ylabel, fontsize=20)
    ax.tick_params(axis="both", which="major", labelsize=18)

    ax.scatter(
        x_raw,
        y_raw,
        color="royalblue",
        s=33,
    )

    ax.plot(rslt["gridpoints"], rslt["curvest"], color="orange", linewidth=3)

    bw = rslt["bandwidth"]
    blue_patch = mpatches.Patch(color="royalblue", label="Raw data")
    orange_patch = mpatches.Patch(
        color="orange", label=f"Local polynomial smooth. Bandwidth = {bw:.3f}"
    )

    plt.legend(
        handles=[blue_patch, orange_patch],
        prop={"size": 16},
    )

    if isinstance(save_as, str):
        plt.savefig(save_as, dpi=300)

    plt.show()



"""Methods for preparing input data."""
from typing import Optional, Tuple, Union

import numpy as np
import pandas as pd


def sort_by_x(
    data: Union[np.ndarray, pd.DataFrame], xcol: Union[int, str] = 0
) -> Union[np.ndarray, pd.DataFrame]:
    """Sort input data by x column."""
    if isinstance(data, pd.DataFrame):
        if isinstance(xcol, int):
            xcol = list(data)[xcol]
        data.sort_values(by=xcol, ascending=True, inplace=True)
    else:  # np.ndarray
        data = data[np.argsort(data[:, xcol])]

    return data


def process_inputs(
    x: Union[np.ndarray, pd.Series],
    y: Union[np.ndarray, pd.Series],
    derivative: int,
    degree: Optional[int],
    gridsize: Optional[int],
    a: Optional[float],
    b: Optional[float],
) -> Tuple[np.ndarray, np.ndarray, int, int, float, float]:
    """Process input arguments for func locpoly."""
    # Turn x (predictor) and y (response variable) into np.ndarrays
    if isinstance(x, pd.Series):
        x, y = np.asarray(x), np.asarray(y)

    if degree is None:
        degree = derivative + 1

    if gridsize is None:
        if len(x) > 400:
            gridsize = 401
        else:
            gridsize = len(x)

    if a is None:
        a = min(x)

    if b is None:
        b = max(x)

    return x, y, degree, gridsize, a, b


def get_example_data() -> pd.DataFrame:
    """Load motorcycle data."""
    # Already sorted by x
    return pd.read_stata("kernreg/example_data/motorcycle.dta")



"""Functions to jit.

The following functions are imported by the main modules
locpoly and linear_binning. There, numba.njit will be applied.
"""
import math
from typing import Tuple

import numpy as np


def include_weights_from_endpoints(
    xcounts: np.ndarray,
    ycounts: np.ndarray,
    y: np.ndarray,
    xgrid: np.ndarray,
    gridsize: int,
) -> Tuple[np.ndarray, np.ndarray]:
    """Attach weight from values outside the grid to corresponding endpoints."""
    for index, value in enumerate(xgrid):
        if value < 1:
            xcounts[0] += 1
            ycounts[0] += y[index]
        elif value >= gridsize:
            xcounts[gridsize - 1] += 1
            ycounts[gridsize - 1] += y[index]

    return xcounts, ycounts


def is_sorted(a: np.ndarray) -> bool:
    """Checks if the input array is sorted ascendingly."""
    for i in range(a.size - 1):
        if a[i + 1] < a[i]:
            return False
    return True


def get_kernelweights(bandwidth: float, delta: float) -> np.ndarray:
    """Compute approximated weights for the Gaussian kernel.

    Arguments:
        bandwidth: Bandwidth.
        delta: Bin width.

    Returns:
        weights: Kernel weights.

    """
    # tau is chosen so that the interval [-tau, tau] is the
    # "effective support" of the Gaussian kernel,
    # i.e. K is effectively zero outside of [-tau, tau].
    # According to Wand (1994) and Wand & Jones (1995), tau = 4 is a
    # reasonable choice for the Gaussian kernel.
    tau = 4

    # Note that L < N, where N is the total number of observations.
    L = math.floor(tau * bandwidth / delta)
    length = 2 * L + 1

    weights = np.zeros(length)

    # Determine midpoint of weights
    mid = L + 1

    # Compute the kernel weights
    for j in range(L + 1):

        # Note that the mid point (weights[mid - 1]) receives a weight of 1.
        weights[mid - 1 + j] = math.exp(-((delta * j / bandwidth) ** 2) / 2)

        # Because of the kernel's symmetry, weights in equidistance
        # below and above the midpoint are identical.
        weights[mid - 1 - j] = weights[mid - 1 + j]

    return weights


def combine_bincounts_kernelweights(
    xcounts: np.ndarray,
    ycounts: np.ndarray,
    weights: np.ndarray,
    degree: int,
    gridsize: int,
    bandwidth: float,
    binwidth: float,
) -> Tuple[np.ndarray, np.ndarray]:
    r"""Combine bin counts and bin averages with kernel weights.

    Use direct convolutions to compute binned
    approximations to :math:`X'W X` and :math:`X'W y`,
    denoted by ``x_weighted`` and ``y_weighted``.

    Arguments:
        xcounts: Binned x-values ("bin counts") of length ``gridsize``.
        ycounts: Binned y-values ("bin averages") of length ``gridsize``.
        weights: Approximated weights for the Gaussian kernel.
        degree: Degree of the polynomial.
        gridsize: Number of equally-spaced grid points.
        bandwidth: Bandwidth.
        binwidth: Bin width.

    Returns:
        x_weighted: Binned approximation to :math:`X'W X`.
        y_weighted: Binned approximation to :math:`X'W y`.

    """
    tau = 4
    L = math.floor(tau * bandwidth / binwidth)
    length = 2 * L + 1
    mid = L + 1

    colx = 2 * degree + 1
    coly = degree + 1

    x_weighted = np.zeros((gridsize, colx))
    y_weighted = np.zeros((gridsize, coly))

    for g in range(gridsize):
        if xcounts[g] != 0:
            for i in range(max(0, g - L - 1), min(gridsize, g + L)):

                if 0 <= i <= gridsize - 1 and 0 <= g - i + mid - 1 <= length - 1:
                    fac_ = 1.0

                    x_weighted[i, 0] += xcounts[g] * weights[g - i + mid - 1]
                    y_weighted[i, 0] += ycounts[g] * weights[g - i + mid - 1]

                    for j in range(1, colx):
                        fac_ *= binwidth * (g - i)

                        x_weighted[i, j] += xcounts[g] * weights[g - i + mid - 1] * fac_

                        if j < coly:
                            y_weighted[i, j] += (
                                ycounts[g] * weights[g - i + mid - 1] * fac_
                            )

    return x_weighted, y_weighted



"""Implementation for the linear binning procedure."""
from typing import Tuple

from numba import njit
import numpy as np

from kernreg.funcs_to_jit import include_weights_from_endpoints


include_weights_from_endpoints_jitted = njit(include_weights_from_endpoints)


def linear_binning(
    x: np.ndarray,
    y: np.ndarray,
    gridsize: int,
    a: float,
    binwidth: float,
    truncate: bool = True,
) -> Tuple[np.ndarray, np.ndarray]:
    r"""Apply linear binning to x and y.

    Linear binning generates bin counts (x-dimension) and bin averages
    (y-dimension) over an equally spaced grid.
    In essence, bin counts are obtained by assigning the raw data to
    neighboring grid points. A bin count represents the
    amount of data in the neighborhood of its corresponding grid point.
    Counts on the y-axis display the respective bin averages.

    The linear binning strategy is based on the transformation
    ``xgrid[i] = ((x[i] - a) * delta) + 1``,
    which maps each ``x[i]`` onto its corresponding gridpoint.
    The integer part of ``xgrid[i]`` indicates the two nearest bin centers to ``x[i]``.
    Additionally, the "fractional part"
    or "bin weight" is computed, which contains the weight attached to the
    two nearest bin centers:

    ``binweights = xgrid - bincenters``.

    The corresponding weights are ``1 - binweights`` for the bin to the left and
    ``binweights`` for the bin to the right.

    If ``truncate`` is True, end observations are trimmed.
    Otherwise, weight from end observations is given to corresponding
    end grid points.


    Arguments:
        x: Array of the predictor variable. Shape (N,). Missing values are not accepted.
            Must be sorted ascendingly.
        y: Array of the response variable. Shape (N,). Missing values are not accepted.
            Must come pre-sorted by ``x``.
        gridsize: Number of equally-spaced grid points in the ``x``-dimension.
            Over this grid, ``x`` and ``y`` are binned.
        a: Start point of the grid.
        binwidth: Bin width.
        truncate: If True, trim endpoints.

    Returns:
        xcounts: Array of binned x-values ("bin counts"). Of length ``gridsize``.
        ycounts: Array of binned y-values ("bin averages"). Of length ``gridsize``.

    """
    N = len(x)

    xcounts = np.zeros(gridsize)
    ycounts = np.zeros(gridsize)
    xgrid = np.zeros(N)
    binweights = np.zeros(N)
    bincenters = [0] * N

    # Map x into set of corresponding grid points
    for i in range(N):
        xgrid[i] = ((x[i] - a) / binwidth) + 1

        # The integer part of xgrid indicates the two nearest bin centers to x[i]
        bincenters[i] = int(xgrid[i])
        binweights[i] = xgrid[i] - bincenters[i]

    for point in range(gridsize):
        for index, value in enumerate(bincenters):
            if value == point:
                xcounts[point - 1] += 1 - binweights[index]
                xcounts[point] += binweights[index]

                ycounts[point - 1] += (1 - binweights[index]) * y[index]
                ycounts[point] += binweights[index] * y[index]

    if truncate is False:
        xcounts, ycounts = include_weights_from_endpoints_jitted(
            xcounts, ycounts, y, xgrid, gridsize
        )

    return xcounts, ycounts



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['kernreg']

package_data = \
{'': ['*'], 'kernreg': ['example_data/*']}

install_requires = \
['matplotlib>=3.3.3,<4.0.0',
 'mypy-extensions>=0.4.3,<0.5.0',
 'numba>=0.52.0,<0.53.0',
 'numpy>=1.19.5,<2.0.0',
 'pandas>=1.2.0,<2.0.0']

setup_kwargs = {
    'name': 'kernreg',
    'version': '0.0.8',
    'description': 'Tool for non-parametric curve fitting using local polynomials.',
    'long_description': '# KernReg\n[![PyPI](https://img.shields.io/pypi/v/kernreg.svg)](https://pypi.org/project/kernreg/)\n[![Continuous Integration](https://github.com/segsell/kernreg/workflows/Continuous%20Integration/badge.svg?branch=main)](https://github.com/segsell/kernreg/actions?query=workflow%3A%22Continuous+Integration%22)\n[![Codecov](https://codecov.io/gh/segsell/kernreg/branch/main/graph/badge.svg)](https://codecov.io/gh/segsell/kernreg)\n[![Codacy Badge](https://app.codacy.com/project/badge/Grade/5dd752959ec8415c8fa9cc9c18ac7d9a)](https://www.codacy.com/gh/segsell/kernreg/dashboard?utm_source=github.com&amp;utm_medium=referral&amp;utm_content=segsell/kernreg&amp;utm_campaign=Badge_Grade)\n[![Black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\n## Introduction\n**KernReg** provides a pure-Python routine for local polynomial kernel regression based on [Wand & Jones (1995)](http://matt-wand.utsacademics.info/webWJbook/) and their accompanying *R* package [KernSmooth](https://www.rdocumentation.org/packages/KernSmooth/versions/2.23-18). In addition, **KernReg** comes with an automatic bandwidth selection procedure that minimizes the residual squares criterion proposed by [Fan & Gijbels (1996)](https://www.taylorfrancis.com/books/local-polynomial-modelling-applications-fan-gijbels/10.1201/9780203748725).\n\n**KernReg** allows for the estimation of a regression function as well as their *v*th derivatives. The degree of the polynomial *p* must be equal to ```v + 1```,\n```v + 3```, ```v + 5```, or ```v + 7```.\n\n<p align="center">\n  <img width="650" height="450" src="https://github.com/segsell/hypermodern-kernreg/blob/main/docs/images/Arthur_Radebaugh_retrofuturism.jpg?raw=true">\n</p>\n\nLocal polynomial fitting provides a simple way of finding a functional relationship between two variables (where X typically denotes the predictor, and Y the response variable)  without the imposition of a parametric model. It is a natural extension of local mean smoothing, as described by [Nadaraya (1964)](https://www.semanticscholar.org/paper/On-Estimating-Regression-Nadaraya/05175204318c3c01e3301fd864553071039605d2#paper-header) and [Watson (1964)](http://www.jstor.org/stable/25049340). Instead of fitting a local mean, local polynomial smooting involves fitting a local *p*th-order polynomial via locally weighted least-squares. The Nadaraya–Watson estimator is thus equivalent to fitting a local polynomial of degree zero. Local polynomials of higher order have better bias properties and, in general, do not require bias adjustment at the boundaries of the regression space.\n\n## Installation\nInstall **KernReg** via PyPI.\n\n```console\n$ pip install kernreg\n```\n\n## Quick-Start\n```python\nimport kernreg as kr\n\nmotorcycle = kr.get_example_data()\nx, y = motorcycle["time"], motorcycle["accel"]\n\n# By default, only x and y need to be provided.\n# Derivative = 0 is chosen by default\n# and hence the polynomial degree = 0 + 1.\nrslt_default = kr.locpoly(x, y)\nkr.plot(x, y, rslt_default, "motorcycle_default_fit.png")\n```\n![default fit](https://github.com/segsell/kernreg/blob/main/docs/images/motorcycle_default_fit.png?raw=true)\n\n```python\n# We can improve on the default specification by\n# choosing a higher order polynomial\nrslt_user = kr.locpoly(x, y, degree=3)\nkr.plot(x, y, rslt_user, "motorcycle_user_fit.png")\n```\n![user fit](https://github.com/segsell/kernreg/blob/main/docs/images/motorcycle_user_fit.png?raw=true)\n\n## References\nFan, J. and Gijbels, I. (1996). [Local Polynomial Modelling and Its Applications](https://www.taylorfrancis.com/books/local-polynomial-modelling-applications-fan-gijbels/10.1201/9780203748725). *Monographs on Statistics and Applied Probability, 66*. Chapman & Hall.\n\nWand, M.P. & Jones, M.C. (1995). [Kernel Smoothing](http://matt-wand.utsacademics.info/webWJbook/). *Monographs on Statistics and Applied Probability, 60*. Chapman & Hall.\n\nWand, M.P. and Ripley, B. D. (2015). [KernSmooth:  Functions for Kernel Smoothing for Wand and Jones (1995)](http://CRAN.R-project.org/package=KernSmooth). *R* package version 2.23-18.\n\n-----\n`*` The image is taken from futurist illustrator [Arthur Radebaugh\'s (1906–1974)](http://www.gavinrothery.com/my-blog/2012/7/15/arthur-radebaugh.html)\nSunday comic strip *Closer Than We Think!*, which was published by the Chicago Tribune - New York News Syndicate from 1958 to 1963.\n',
    'author': 'Sebastian Gsell',
    'author_email': 'sebastian.gsell93@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/segsell/kernreg',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.1,<3.9',
}


setup(**setup_kwargs)
